#!/usr/local/bin/perl
#
#    wwgetall ver3.xx
#         by Yoshioka Tsuneo
#         welcome any email to: QWF00133@nifty.ne.jp
#         Copy,Edit,Re-distribute FREE
#  URL: http://infonet.aist-nara.ac.jp/~tsuneo-y/soft/wwgetall/wwgetall.html
#
use Socket;
use Fcntl;            

$Version = "3.14";
$wwgetallHomePage = "http://infonet.aist-nara.ac.jp/~tsuneo-y/soft/wwgetall/wwgetall.html";

# use POSIX; # for EINPROGRESS 
# not exist POSIX.pm on Perl for Win32
eval('require "POSIX.pm"; $EINPROGRESS = POSIX::EINPROGRESS(),"\n"');


local (%filestate);
local (@filehandle);
local (%checkedURL);

$USER_AGENT = "User-Agent: wwgetall/$Version; $wwgetallHomePage\r\n";

$MAXSOCKET = 5;		# max parallel connections.
#$MAXCONNECTPERHOST = 5;		# max connections per server host.(not implemented now)

local($socketnum)=0;			# current connections.
local($htmldir) = "";			# recieved http object store path

local($proxy_host);		# proxy host name
local($proxy_port);		# proxy port number

local($SH_BASE) = "SH00000000";	# Socket Handler  (plused by connection)
local($FH_BASE) = "FH00000000";	# File Handler    (plused by conenction)

$SIG{'PIPE'} = sub {print "SIGPIPE happen\n";};	# for connection shutdown

$matchurl = "";			# regular expression for decide get URL.
$fullbody=0;			# if =1, Save Response Header too.
local(@URLQUEUE);		# URL Array for request.
local(%hostconnectnum);
local($KEEPALIVE) = 1;		# Do persistent connection(1) or not(0),

$MAX_CONTENT_LENGTH = 1000000000;# Content-Length-Limit size

#  read default setting file "$HOME/.wwgetall"
$HOME = ($ENV{'HOME'} || ".");
if($HOME !~ m|/$|){$HOME .= "/";}
if( -r "${HOME}.wwgetall"){
    require "${HOME}.wwgetall";
}

sub min
{
	local($a,$b) = @_;
	if($a<$b){return $a;}else{return $b}
}
sub max
{
	($_[0] > $_[1]) ? $_[0] : $_[1];
}
#  mkdir recursive (mkdir -p)
sub mkdirp
{
    local($dir)=@_;
    local(@dirs)=split(/\//,$dir);
    local($dir2,$dir3);
    foreach $dir2(@dirs){
	$dir3 .= ($dir2 . "/");
	mkdir $dir3,0777;
    }
}
sub parse_url
{
    local($url) = @_;
    local($proto,$host,$port,$path);

    if($url =~ m|(http\|ftp)://([^:/]+)(:(\d+))?([^\#\?]*)|){
	$proto = $1;
	$host = $2;
	$port = $4 || 80;
	$path = $5 || "/";
#	if($path =~ m|/$|){$path .= "index.html";}
    }else{
	return 0;
    }
    return ($proto,$host,$port,$path);
}
sub build_url
{
    local($proto,$host,$port,$path) = @_;
    local($url);

#    if($path =~ m|/$|){$path .= "index.html";}
    $url = $proto . "://" . $host . (($port==80) ? "" : ":$port") . $path;
    return $url;
}
sub clean_url
{
    local($url)=@_;

    local($proto,$host,$port,$path) = &parse_url($url);
    if(! $proto){return 0;}
    $url = &build_url($proto,$host,$port,$path);
    $_[0] = $url;
    return 1;
}
sub url2urlfname
{
    local($url) = @_;
    $url =~ s|/$|/index.html|;
    return $url;
}

sub check_existfcntl
{	# fcntl check (win32 don't have fcntl)
    local(*SH);
    local($existfcntl);

    socket(SH,AF_INET,SOCK_STREAM,0) || die "can't open socket:$!";
    eval('fcntl(SH,F_SETFL,O_NONBLOCK)');
    if($@){
	warn "Warning: Perhaps i'ts WIN32. fcntl() failed.can't use non-blocking connect():$!";
	$existfcntl=0;
    }else{
	$existfcntl=1;
    }
    close(SH);
    return $existfcntl;
}

# from WWWcp
sub base64encode_string
{
	local($arg) = @_;
	local($b64) = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
	local($len,$rem,$bin,$code,$char);

	$len  = length($arg)*8;
	$bin = unpack("B".$len, $arg);
	$rem = $len % 6;
	$bin .= "0" x (6-$rem) if $rem;
	$code = "";
	while(length($bin)){
		$char = substr($bin,0,6);
		$bin = substr($bin,6);
		$code .= substr($b64,unpack("C",pack("B8","00$char")),1);
	}
	$rem = length($code)%4;
	$code .= "=" x (4-$rem) if $rem;
	return 	$code;
}
sub usage
{
print "wwgetall ver $Version\n";
    print <<'_EOT_';
*********************************************************************
    wwgetall ver 3.xx  --recursive http object collector--
	 nonblocking parallel connection,Keep-Alive support.
	by Yoshioka Tsuneo(QWF00133@nifty.ne.jp)/Welcome any e-mail
 URL: http://infonet.aist-nara.ac.jp/~tsuneo-y/wwgetall/wwgetall.html
*********************************************************************
_EOT_
    print <<'_EOT_';
wwgetall [options] [<URL>|@<URLLISTFILENAME>]
  options:
    --proxy=<host>:<port>   ;# use proxy server.
    --rootdir=<dir>	;# directory to save retrieved files.(default:current)
    --log=<filename>	;# Log File Name(default:rootdir/wwgetall.log)
    --match=<url>	;# regular expression that define retvieve URLs.
	;# (default: "^(?i)(<URL directory>)(\.s?html?|\.txt|\.gif|\.jpe?g|\.rdf|\.css)$")
    --fullbody		;# save not only http response body but also response header.
    --connection=<pararel connection number> 
                        ;# nonblocking pararell connections.(default:5) 
    --no-keepalive	;# not use Connection: Keep-Alive.
    --htmlonly		;# retvieve html file only.(can't use with --match)
    --alltype		;# retvieve all type of file. (can't use with --match)
    --check-modified (--ims) ;# send If-Modified-Since header.
                  ;#  this option set "--fullbody" option automatically.
		  ;# previous retvieval must contain "--fullbody" option.
    -1			;# Not Recursive
    -D<num>		;# set debug level. 
    --auth=ID:PASSWORD  ;# Authorization
    -U			;# Fake User-Agent as Mozilla
  examples:
    wwgetall http://infonet.aist-nara.ac.jp/ --rootdir=/var/tmp --htmlonly

_EOT_
}

##################MAIN###################################
if($0 eq __FILE__){
    local($i,$url);

    while($_ = $ARGV[0]){
	if(/^-/){
	    if(/^--rootdir=(\S+)/){
		$htmldir = $1;
		if($htmldir !~ m|/$|){$htmldir.="/";}
	    }elsif(/^--proxy=(\S+):(\d+)$/){
		$proxy_host = $1;
		$proxy_port = $2;
		$use_proxy = 1;
	    }elsif(/^--log=(\S+)/){
		$logfile = $1;
	    }elsif(/^--match=(\S+)$/){
		$matchurl=$1;
	    }elsif(/^--fullbody/){
		$fullbody=1;
	    }elsif(/^--connection=(\d+)/){
		$MAXSOCKET=$1;
	    }elsif(/^--no-keepalive/){
		$KEEPALIVE = 0;
	    }elsif(/^--htmlonly/){
		$htmlonly = 1;
	    }elsif(/^--alltype/){
		$alltype = 1;
	    }elsif(/^(--check-modified|--ims)/){
		$check_modified=1;
		$fullbody=1;
	    }elsif(/^-1/){
		$matchurl = '^$';#'
	    }elsif(/^-D(\d+)/){
		$debug = $1;
	    }elsif(/^-U/){
		$USER_AGENT = "User-Agent: Mozilla/4.0 (compatible; MSIE 4.01; wwgetall/$Version; $wwgetallHomePage)\r\n";
	    }elsif(/^--auth=([^:]*:.*)$/i){
		$AuthStr = $1;
	    }else{
		usage();
		exit 1;
	    }
	}elsif(/^\@(.*)/){
	    local(*FH);
	    print "Loading $1 ...\n";
	    open(FH,"<$1") || die "can't open [$1]:$!";
	    while(<FH>){
		tr/\r\n//d;
		if(/^.+/){
		    push(@ARGV,$_);
		}
	    }
	    close(FH);
	    print "Loaded! \n";
	}else{
	    &clean_url($_) || die "Illegal URL[$_]";
	    if(! $checkedURL{&url2urlfname($_)}){
		push(@URLQUEUE,$_);
		$checkedURL{&url2urlfname($_)} = 1;
	    }
	}
	shift;
    }
    if(! @URLQUEUE){
	&usage;
	exit 1;
    }

    if($matchurl eq ""){
	local($url);
	$matchurl = "";
	$matchurl .= "^(?i)(";
	foreach (@URLQUEUE){
	    $url = $_;
	    if($url ne $URLQUEUE[0]){
		$matchurl .= "|";
	    }
	    $url =~ s/[^\/]+$//;
	    $matchurl .= "$url" . ".*";
	}
	if($htmlonly){
	    $matchurl .= ')\.s?html?$'; # '
	}elsif($alltype){
	    $matchurl .= ')'; # '	    
	}else{			# default 
	    # $matchurl .= ')\.html?|(\.gif|\.jpe?g|\.rdf|\.css)$'; # '
	    $matchurl .= ')(\.s?html?|\.txt|\.gif|\.jpe?g|\.rdf|\.css)$'; # '
	}
    }
    local($oldfh)=select(STDOUT);$|=1;select($oldfh);
    local($oldfh)=select(STDERR);$|=1;select($oldfh);
     if($logfile eq ""){
	$logfile = "${htmldir}wwgetall.log";
	open(LOG,">$logfile")||die "can't open logfile [$logfile]:$!";
	local($oldfh)=select(LOG);$|=1;select($oldfh);
    }
    if($debug){ 
	print "Waiting for Replay[$matchurl]...\n";
    }
    $existfcntl = &check_existfcntl;

    study $matchurl;
    mkdir($htmldir,0777);
    &process_URLQUEUE(0);

    while(1){
	if($rin =~ /^\0*$/ && $win =~ /^\0*$/){
	    if($debug){print "All Connection is done.\n";}
	    last;
	}
#	print "Waiting for Reply...(connection num = $socketnum)\n";
	$ret = select($rout=$rin,$wout=$win,undef,undef);
	if($ret<=0){
		local($i);
		print "--------------------------\n";
		for($i=0;$i<length($rin)*8;$i++){
			print("vec(rin,$i,1)=",vec($rin,$i,1),"\n");
			print("vec(rout,$i,1)=",vec($rout,$i,1),"\n");
		}
		for($i=0;$i<length($win)*8;$i++){
			print("vec(win,$i,1)=",vec($win,$i,1),"\n");
			print("vec(wout,$i,1)=",vec($wout,$i,1),"\n");
		}
		print "--------------------------\n";
		die "select failed:$!";
	}
	for($i=0;$i<length($rin)*8;$i++){
	    if(vec($rout,$i,1)==1){
		&check_filehandle($i,"r");
	    }
	}
	for($i=0;$i<length($win)*8;$i++){
	    if(vec($wout,$i,1)==1){
		&check_filehandle($i,"w");
	    }
	}
    }
    close(LOG);
#    print "---------left URLQUEUE-------\n";
#    print join(",",@URLQUEUE),"\n";
    
}
#########################################################

sub url2fname
{
    local($url) = @_;
    $url =~ s|://|/|;
    $url =~ s|/$|/index.html|;
    return $htmldir . $url;
}
sub process_URLQUEUE
{
    local($keepalive,$fd) = @_;
    local($KEEPSH,$keephost);
    local($url);
    local($i);
    if($keepalive){
	$KEEPSH = $sockethandle[$fd];
	local($url) = $urllist[$fd];
	local($proto,$host,$port,$path) = &parse_url($url);
	$keephost = $host;
    }
#    print "URLQUEUE=",join(",",@URLQUEUE),"\n";
    while(scalar(@URLQUEUE)>0 
	  && (($keepalive && ($socketnum <= $MAXSOCKET)) || ($socketnum<$MAXSOCKET))){
	local($server_host,$server_port);

	$url = shift(@URLQUEUE);

	($proto,$host,$port,$path) = &parse_url($url);
	if(! $proto){next;}
	#if($hostconnectnum{$host} >= $MAXCONNECTPERHOST){
	#}
	$url = &build_url($proto,$host,$port,$path);



	if($use_proxy){
	    $server_host = $proxy_host;
	    $server_port = $proxy_port;
	}else{
	    $server_host = $host;
	    $server_port = $port;
	}
	$SH = ++$SH_BASE;
	if($KEEPALIVE && $keepalive){
	    if($debug){print "keepalive=[$keepalive],KEEPSH=[$KEEPSH],keephost=[$keephost],host=[$host],fd=[",fileno($KEEPSH),"]\n";}
	    if($keephost eq $host){
		local($rin,$rout);
		$rin = '';
		vec($rin,fileno($KEEPSH),1)=1;
		local($ret) = select($rout=$rin,undef,undef,0.001);
		if($ret==1){
		    if($debug){print "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n";}
		    &closefd(fileno($KEEPSH),0);
		}else{
		    if($debug){print "**************************************************************\n";}
		    $SH = $KEEPSH;
		    print "[$SH]Connect to [$url] by KeepAlive.\n";
		}
		$keepalive = 0;
	    }
	}
	if($SH eq $SH_BASE){
	    if($socketnum >= $MAXSOCKET){unshift(@URLQUEUE,$url);last;}
	    socket($SH,PF_INET,SOCK_STREAM,0) || die "can't open socket:$!";
	    if($existfcntl){
		fcntl($SH,F_SETFL,O_NONBLOCK) || die "can't set nonblock:$!";
	    }
	    #print "starting inet_aton($server_host)...\n";
	    local($addr) = inet_aton($server_host);
	    #print "end inet_aton($server_host)...\n";
	    if($addr eq ''){
		print "ERROR: inet_aton($server_host)(url=$url)\n";
		print LOG  time() , ".000 " , 0  , " " ,"localhost",' ERR_INET_ATON/','000',' ',0,' ','GET ',$url,' - NONE ','-',"\n";
		next;
	    }
	    if(! connect($SH,sockaddr_in($server_port,$addr))){
		if($! != $EINPROGRESS){
		    print  "ERROR: can't connect to [$server_host:$server_port]($url):$!";
		    print LOG  time() , ".000 " , 0  , " " ,"localhost",' ERR_CONNECT/','000',' ',0,' ','GET ',$url,' - NONE ','-',"\n";
		    next;
		}
	    }
	    $socketnum++;
	    print "[$SH]Connect to [$url].\n";
	}
	$i=fileno($SH);
	$filestate[$i] = "CONNECTING";
	$sockethandle[$i] = $SH;
	$filestarttime[$i] = time();
	$filehandle[$i] = '';
	$filehandle[$i] = '';
	$inputbuf[$i]='';
	$urllist[$i] = $url;
	$last_modified_time[$i] = '';
	# push(@{$HOST2URLLIST{$host}},$url);
	$content_type[$i] = "";
	$hostconnectnum{$host} += 1;
	vec($win,$i,1)=1;	# Wait for Write

	if($debug){print "connecting:[fd=$i][conn=$socketnum][$SH][$url]\n";}
	if($SH ne $SH_BASE){	# keep-alive
	    &check_filehandle($i,'w');
	}
    }
    if($keepalive){&closefd(fileno($KEEPSH),0);}
}
sub check_filehandle($$)
{
    local($i,$mode) = @_;
    local($SH,$FH,$url,$elapsed);
    # print "[fd=$i][conn=$socketnum][mode=$mode][$filehandle[$i]][$filestate[$i]]\n";

    $SH = $sockethandle[$i];
    $FH = $filehandle[$i];
    $url = $urllist[$i];
    $elapsed = time() - $filestarttime[$i];

    if($filestate[$i] eq "CONNECTING"){
	print "Send Request Header to [$urllist[$i]]...\n";
	local($path);		# GET $path HTTP/1.0...
	local($cmd);		# request header
	if(! $use_proxy){
	    (undef,undef,undef,$path) = &parse_url($url);
	}else{
	    $path = $url;
	}
	$cmd = "GET $path HTTP/1.0\r\n";
	print "cmd=$cmd\n";
	if($KEEPALIVE){$cmd .= "Connection: Keep-Alive\r\n";}
	if($check_modified){
# Save LastModifiedTime & ContentType from old file.
	    local(*FH);
	    local($fname)=&url2fname($url);
	    local($last_modified_time)="";
	    if(open(FH,"<${fname}.head")){
		$_ = <FH>;
		if(m|^HTTP/\S+\s+200|){
		    while(<FH>){
			tr/\r\n//d;
			if(/^$/){break;}
			if(/^Last-Modified: (.*)/i){
			    $last_modified_time = $1;
			    print "Last-Modifed at [$last_modified_time]\n";
			}elsif(/^Content-Type: (\S+)/i){
			    $content_type[$i] = $1;
			}
		    }
		}else{
		    print "warning: [$fname] isn't contain 'HTTP/?.? 200'\n";
		}
		close(FH);
	    }else{
		print "warning: can' find response header file [${fname}.head]. to make response header file, use --fullbody option.\n";
	    }
	    if($last_modified_time){
		$cmd .= "If-Modified-Since: $last_modified_time\r\n";
	    }
	}
	$cmd .= $USER_AGENT;
	if($AuthStr){ 
		$cmd .= "Authorization: Basic " . &base64encode_string($AuthStr) . "\r\n";
	}
	
	$cmd .= "\r\n";
	
	# print "cmd=$cmd\n";
	$ret = syswrite($SH,$cmd,length($cmd));
	if(! $ret){
	    print "ERROR: can't write($url):$!\n";
	    print LOG  time() , ".000 " , sprintf("%6d",$elapsed*1000)  , " " ,"localhost",' ERR_CONNECT_WRITE/','000',' ',0,' ','GET ',$url,' - NONE ','- ',$!,"\n";
	    &closefd($i,0);
	    &process_URLQUEUE(0);
	}
	vec($win,$i,1)=0;vec($rin,$i,1)=1; # Wait for Read
	$filestate[$i] = "GETTING_HEAD";
    }elsif($filestate[$i] eq "GETTING_HEAD"){
	local($buf);

	$ret = sysread($SH,substr($inputbuf[$i],length($inputbuf[$i])),10000);
	$buf = $inputbuf[$i];
	# $ret = sysread($SH,$buf,10000);
	# $filebody[$i] .= $buf;

	if($ret<=0){
	    print "ERROR: sysread() = $ret:$!\n";
	    print LOG  time() , ".000 " ,  sprintf("%6d",$elapsed*1000) , " " ,"localhost",' ERR_READ_HEAD/','000',' ',0,' ','GET ',$url,' - NONE ','- ',$!,"\n";
	    &closefd($i,0);
	    &process_URLQUEUE(0);
	    return 1;
	}
	if($buf !~ /\r?\n\r?\n/){
	    return 1;
	}
	#    print "ERROR: can't found double return\n";
	#    print LOG  time() , ".000 " , sprintf("%6d",$elapsed*1000)  , " " ,"localhost",' ERR_HEAD_DOUBLERETURN/','000',' ',0,' ','GET ',$url,' - NONE ','-',"\n";
	#    &closefd($i);&process_URLQUEUE(0);return 1;
	#}
	$filehead[$i] = $head = $`;
	$body = $';
	# $filebody[$i] = $body;
	$filebodylength[$i] = length($body);
	if($head !~ m|^HTTP/\S+\s+(\d+)|){
	    print "Error: Response is illegal. [$head]\n";
	    print LOG  time() , ".000 " , sprintf("%6d",$elapsed*1000)  , " " ,"localhost",' ERR_HEAD_FIRST/','000',' ',0,' ','GET ',$url,' - NONE ','-',"\n";
	    &closefd($i,0);
	    &process_URLQUEUE(0);
	    return 1;
	}
	local($code) =  $1;
	if($code == 304){
	    print "Not Modified.[$urllist[$i]].\n";
	    local($type) = $content_type[$i];
	    &closefd($i,1);
	    &process_URLQUEUE(1,$i);
	    if($type eq "text/html"){
		# print "type=[$type],url=[$url]\n";
		&getlinks($url,&url2fname($url));
	    }
	    &process_URLQUEUE(0);
	    return 1;
	}
	if($head =~ /(^|\n)Content-Type: (\S+)/i){
	    $content_type[$i] = $2;
	}else{
	    $content_type[$i]="";
	}
	# print "content_type = [$content_type[$i]]\n";
	if($head =~ /(^|\n)Content-Length: (\d+)/i){
	    $content_length[$i] = &min($2,$MAX_CONTENT_LENGTH);
	}else{
	    $content_length[$i]=$MAX_CONTENT_LENGTH;
	    #&closefd($i);
	}
	{
	    $fname = &url2fname($url);
	    $FH = ++$FH_BASE;
	    if(-d $fname){rmdir $fname;} mkdirp($fname);rmdir("$fname");
	    if(! open($FH,">$fname")){
		warn "#########Error: can't open $fname:$!";
		closefd($i,0);
		&process_URLQUEUE(0);
		return 1;
	    }
	    binmode($FH);
	}
	$filehandle[$i] = $FH;
	print $FH $body;
	print "Get Response Header From [$url].($filebodylength[$i]/$content_length[$i])\n";
	if($content_length[$i] == $filebodylength[$i]){goto sessioncomplete;}
	$filestate[$i] = "GETTING_BODY";
    }elsif($filestate[$i] eq "GETTING_BODY"){
	local($buf);
	$ret = sysread($SH,$buf,10000);
	if($ret>0){
	    $filebodylength[$i] += $ret;
	    print $FH $buf;
	    print "Recieved Body from [$url].($filebodylength[$i]/$content_length[$i])\n"
	}
	# print "content_length= $content_length[$i],filebody length =",$filebodylength[$i],"\n";
	if($ret<=0 ||  $filebodylength[$i] >= $content_length[$i]){
sessioncomplete:
	    local($ti) = time() - $filestarttime[$i];
	    print "HTTP Session Complete. [$urllist[$i]].($filebodylength[$i])(${ti}sec)\n";
	    if($fullbody){
		local($fname) = &url2fname($url) . ".head";
		print "writing header to [$fname]\n";
		if(open(FH,">$fname")){
		    print FH $filehead[$i];
		    close(FH);
		}else{
		    warn "#######ERROR:can't open [$fname]:$!\n";
		}
	    }
	    local($responsecode)="000";
	    if($filehead[$i] =~ m|^HTTP/\S+\s+(\d+)|){ $responsecode=$1;}
	    # if($filehead[$i] =~ /^[^\r\n]+/){$responsestate = $&;}
	    local($type) = $content_type[$i];
	    local($proto,$host,$port,$path) = &parse_url($url);
	    if($debug){print "gotten[fd=$i][conn=$hostconnectnum{$host}/$socketnum][$SH][$filebodylength[$i]byte][${ti}sec][$urllist[$i]][$filehandle[$i]]\n";}
	    &closefd($i,1);
	    print LOG  time() , ".000 " , sprintf("%6d",$ti*1000) , " " ,"localhost",' /',$responsecode,' ',(-s &url2fname($url)),' ','GET ',$url,' - NONE ',$type,"\n";
	    &process_URLQUEUE(1,$i);
	 #   print "post process_URLQUEUE content_type=[$content_type[$i]]\n";
	    if($type eq "text/html"){
	#	print "Getting Links($url,$htmldir$url)...\n";
		&getlinks($url,&url2fname($url));
	    }
	    &process_URLQUEUE(0);
#	    &process_URLQUEUE(1,$i);
	}
    }
    return 1;
}
sub getlinks
{
    local($srcurl,$fname) = @_;
    local(*FH);

#    print "getlinks($srcurl,$fname) start...\n";
    if(!open(FH,"<$fname")){
	die "can't open [$fname]:$!";
	return;
    }
    binmode(FH);
#    print "srcurl=[$srcurl] fname[$fname] opened\n";
    while(<FH>){
	$line = $_;
start:
	# print "line = [$line]\n";
	if($line !~ s|[^<]*\<|\<|){next;}
	while(($line !~ m|\>|) && ($_ = <FH>)){
	     $line .= $_;
	}
	if($line =~ m|\>|){
	    $lostline = $';
	    $line = $` . ">";
	}else{
	    $lostline ="";
	}
	$line =~ s/\s+/ /g;
	$line =~ s/<\s+/</g;
	$line =~ s/\s+>/>/g;
	$line =~ s/\s*=\s*/=/g;
	$url = '';
	#print "line=[$line],lostline=[$lostline]\n";
#	print "line = [$line]\n";
	if($line =~ m/<a\s([^>]*\s)?href\=(\"([^\"]*)\"|([^\"]\S+))(\s+[^>]*)?>/i){ #"
	    $url = $3 || $4;
	 #   print "matched url=[$url]\n";
	}elsif($line =~ m/<img\s([^>]*\s)?src=(\"([^\"]*)\"|([^\"]\S+))(\s+[^>]*)?>/i){ #"
	    $url = $3 || $4;
	   # print "IMG match=[$url]\n";
	}elsif($line =~ m/<frame\s([^>]*\s)?src=(\"([^\"]*)\"|([^\"]\S+))(\s+[^>]*)?>/i){ #"
	    $url = $3 || $4;
	}elsif($line =~ m/<body\s([^>]*\s)?background=(\"([^\"]*)\"|([^\"]\S+))(\s+[^>]*)?>/i){ #"
	    $url = $3 || $4;
	   # print "Backgournd match = [$url]\n"
	}
	$url =~ s|\#.*$||;	# erase label
	if($url ne ""){
	    # print "1=============srcurl=[$srcurl],url=[$url]\n";
	    $url = &addurl($srcurl,$url);
	    # print "2=========================srcurl=[$srcurl]\n";
	    if($url =~ m|^(http\|ftp)://|){
		local($urlfname) = &url2urlfname($url);
		# print "found url=[$url][$urlfname],matchurl is [$matchurl]\n";
		if($urlfname  =~ /$matchurl/){ 
		  #   print "check url [$url],checkedurl=[$checkedURL{$url}]\n";
		    if(! $checkedURL{$urlfname}){
			print "  found new[$url].\n";
			push(@URLQUEUE,$url);
			$checkedURL{$urlfname} = 1;
		    }
		}
	    }
#	    print "3=========================srcurl=[$srcurl]\n";
	}
	if($lostline ne ""){
	    $line = $lostline;
	    goto start;
	}
    }
    #print "getlinks($srcurl,$fname) end...\n";
    close(FH);
}
sub clean_path
{
    local($srcpath) = @_;
    local($destpath);
    local(@srcpathlist,@destpathlist);

    @srcpathlist = split(/\//,$srcpath);
    if($srcpathlist[0] eq ""){shift(@srcpathlist);}
    if($srcpath =~ m|/$|){push(@srcpathlist,"");}
   # print "srcpathlist=",join(",",@srcpathlist),"\n";
    while(@srcpathlist){
	$_ = shift(@srcpathlist);
	#if(/^$/){
	#    next;
	#}els
        if(/^\.\.$/){
	    pop(@destpathlist);
	}elsif(/^\.$/){
	    ;
	}else{
	    push(@destpathlist,$_);
#	    print "destpathlist=[@destpathlist]\n";
	}
    }
    $destpath = join("/",@destpathlist);
    if($destpath !~ /^\//){$destpath = "/" . $destpath;}
   # print "clean_path():end / srcpath = [$srcpath],destpath=[$destpath]\n";
    return $destpath;
    $_[0] = $destpath;
    return 1;
}
sub addurl
{
    local($srcurl,$addurl) = @_;

#    print "addurl:srcurl=[$srcurl],dest=[$addurl]\n";
    local($newurl,@pathlist);
    local($proto,$host,$port,$path) = &parse_url($addurl);
    if(! $proto){
	if($addurl =~ /:/){return 0;}
	($proto,$host,$port,$path) = &parse_url($srcurl);
	if($addurl !~ m|^/|){
	    @pathlist = split(/\//,$path);
	    if($path =~ m|/$|){push(@pathlist,"");}
	    pop(@pathlist);
	    push(@pathlist,$addurl);
#	    print "pathlist=",join(",",@pathlist),"\n";
	    $path = join("/",@pathlist);
	    if($path eq ""){$path = "/";}
	    $path = &clean_path($path);
	}else{
	    $path = $addurl;
	}
    }
    #print "proto=$proto,host=$host,port=$port,path=$path\n";
    $newurl = &build_url($proto,$host,$port,$path);
    return $newurl;
}
sub closefd
{
    local($fd,$keepalive)=@_;
    local($SH,$FH);
    local($url);

    $SH = $sockethandle[$fd];
    $url = $urllist[$fd];
    if($keepalive!=1){
	close($SH);
	$sockethandle[$fd]='';
	$socketnum -= 1;
	vec($rin,$fd,1)=0;
	vec($win,$fd,1)=0;
	$urllist[$fd] = '';
	$hostconnectnum{$host} -= 1;
    }
    $FH = $filehandle[$fd];
    if($FH){
	close($FH);
	$filehandle[$fd]='';
    }
    # print "closing[$fd]...SH=[$SH],FH=[$FH],keepalive=$keepalive,socketnum=$socketnum\n";
    local($proto,$host,$port,$path) = parse_url($url);
    $filehead[$fd]='';
    $filebody[$fd]='';
#    if($blockinput && $socketnum<$MAXSOCKET){
#	$blockinput = 0;
#	&process_URLQUEUE($keepalive,$SH,$host);
#    }
#    &process_URLQUEUE($keepalive,$SH,$host);
}

